import WUPBaseElement from "./baseElement";
import { parseMsTime } from "./helpers/styleHelpers";
import { WUPcssScrollSmall } from "./styles";
var OpenCases;
(function (OpenCases) {
    /** When $open() is called programmatically */
    OpenCases[OpenCases["onManualCall"] = 0] = "onManualCall";
})(OpenCases || (OpenCases = {}));
var CloseCases;
(function (CloseCases) {
    /** When $close() is called programmatically */
    CloseCases[CloseCases["onManualCall"] = 0] = "onManualCall";
})(CloseCases || (CloseCases = {}));
/** Base modal element */
export default class WUPBaseModal extends WUPBaseElement {
    // #ctr = this.constructor as typeof WUPBaseModal;
    static get $styleRoot() {
        return `:root {
        --modal-anim-t: 400ms;
      }`;
    }
    static get $style() {
        return `${super.$style}
      :host {
        --modal-anim: var(--modal-anim-t) cubic-bezier(0, 0, 0.2, 1) 0ms;
        z-index: 8000;
        display: none;
        opacity: 0;
        position: fixed!important;
        border-radius: var(--border-radius);
        padding: var(--base-margin);
        box-sizing: border-box;
        white-space: pre-line;
        overflow: auto;
        overflow: overlay;
      }
      :host[open] { display: block; }
      :host[show] { opacity: 1; }
      ${WUPcssScrollSmall(":host")}
      @media not all and (prefers-reduced-motion) {
        :host {
          transition: opacity var(--modal-anim), transform var(--modal-anim);
        }
      }`;
    }
    // static $defaults: WUP.BaseModal.Options = {};
    #isOpened = false;
    /** Returns if element is open (before show-animation started)
     * @tutorial Troubleshooting
     * * stack: $open() > `$isOpened:true` > opening > opened
     * * stack: $close() > closing > closed > `$isOpened:false`
     * * to listen to animation-end use events `$open` & `$close` OR methods `$open().then(...)` & `$close().then(... )` */
    get $isOpened() {
        return this.#isOpened;
    }
    #isOpening;
    /** Returns if element is opening (only if animation enabled) */
    get $isOpening() {
        return this.#isOpening === true;
    }
    /** Returns if element is closed (after hide-animation finished)
     * @tutorial Troubleshooting
     * * stack: $open() > `$isOpened:true` > opening > opened
     * * stack: $close() > closing > closed > `$isOpened:false`
     * * to listen to animation-end use events `$open` & `$close` OR methods `$open().then(...)` & `$close().then(... )` */
    get $isClosed() {
        return !this.#isOpened && !this.#isClosing;
    }
    #isClosing;
    /** Returns if element is opening (only if animation enabled) */
    get $isClosing() {
        return this.#isClosing === true;
    }
    /** Fires before opening
     * @tutorial rules
     * * can be prevented via `e.preventDefault()` */
    $onWillOpen;
    /** Fires after element is opened (after animation ends) */
    $onOpen;
    /** Fires before closing
     * @tutorial rules
     * * can be prevented via `e.preventDefault()` */
    $onWillClose;
    /** Fires after element is closed (after animation ends) */
    $onClose;
    _whenOpen;
    /** Open element
     * @returns Promise resolved by animation-end ('true' if it finished & wasn't prevented) */
    $open() {
        if (this.$isReady) {
            try {
                return this.goOpen(0 /* OpenCases.onManualCall */, null);
            }
            catch (err) {
                return Promise.reject(err);
            }
        }
        return new Promise((res, rej) => {
            setTimeout(() => {
                if (!this.$isReady) {
                    rej(new Error(`${this.tagName}. Impossible to show: not appended to document`));
                }
                else {
                    this.goOpen(0 /* OpenCases.onManualCall */, null).then(res);
                }
            }, 1); // 1ms need to wait forReady
        });
    }
    _whenClose;
    /** Close element
     * @returns Promise resolved by animation-end ('true' if it finished & wasn't prevented) */
    $close() {
        return this.goClose(0 /* CloseCases.onManualCall */, null);
    }
    gotReady() {
        super.gotReady();
        this._willFocus && clearTimeout(this._willFocus);
        delete this._willFocus; // prevent default autofocus behavior since here need to use it only on open
    }
    // protected override gotChanges(propsChanged: string[] | null): void {
    //   super.gotChanges(propsChanged);
    // }#
    /** Returns animation time defined according to styles */
    get animTime() {
        return parseMsTime(getComputedStyle(this).transitionDuration);
    }
    /** Open element. @openCase as reason of open() */
    goOpen(openCase, ev) {
        if (this._whenOpen) {
            return this._whenOpen;
        }
        const e = this.fireEvent("$willOpen", { cancelable: true, bubbles: true, detail: { openCase } });
        if (e.defaultPrevented) {
            return Promise.resolve(false);
        }
        // clear state
        this._whenClose = undefined;
        this.#isClosing = undefined;
        this.#isOpened = true;
        this.#isOpening = true;
        // apply open styles
        this.setAttribute("open", ""); // apply display effect
        this.gotOpen(openCase, ev);
        setTimeout(() => {
            this.scroll({ left: 0, top: 0, behavior: "instant" }); // scroll at the top: to clear prev possible state
            this.setAttribute("show", ""); // apply show animation
        }); // timeout to allow animation works
        // wait for animation
        this._whenOpen = new Promise((res) => {
            setTimeout(() => {
                if (!this._whenOpen) {
                    res(false); // possible when call $close during the opening
                    return;
                }
                this.#isOpening = undefined;
                res(true);
                setTimeout(() => this.fireEvent("$open", { cancelable: false, bubbles: true }));
            }, this.animTime);
        });
        return this._whenOpen;
    }
    /** Hide element; @closeCase as reason of close() */
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    goClose(closeCase, ev, immediately) {
        if (this._whenClose) {
            return this._whenClose;
        }
        if (!this._whenOpen) {
            return Promise.resolve(true); // possible when on init $close is called
        }
        const e = this.fireEvent("$willClose", { cancelable: true, bubbles: true, detail: { closeCase } });
        if (e.defaultPrevented) {
            return Promise.resolve(false);
        }
        // clear state
        this._whenOpen = undefined;
        this.#isOpening = undefined;
        this.#isClosing = true;
        // apply animation
        this.removeAttribute("show");
        this.gotClose(closeCase, ev, immediately);
        this._whenClose = new Promise((res) => {
            setTimeout(() => {
                if (!this._whenClose) {
                    res(false); // possible when call $open during the hiding
                    return;
                }
                this.resetState();
                res(true);
                setTimeout(() => this.fireEvent("$close", { cancelable: false, bubbles: true }));
            }, immediately ? 0 : this.animTime);
        });
        return this._whenClose;
    }
    focus() {
        if (!this.$isOpened) {
            return false;
        }
        return super.focus();
    }
    /* Called to clear current state/vars to init: after gotClose end */
    resetState() {
        this.removeAttribute("open");
        this.#isOpened = false;
        this.#isClosing = undefined;
        this.#isOpening = undefined;
        this._whenClose = undefined;
        this._whenOpen = undefined;
    }
    /* Call it to call baseElement.dispose */
    disposeEvents() {
        super.dispose();
    }
    dispose() {
        this.resetState();
        super.dispose();
    }
}
