import WUPBaseElement from "../baseElement";
import WUPFormElement from "../formElement";
import isEqual from "../helpers/isEqual";
import nestedProperty from "../helpers/nestedProperty";
import onFocusLostEv from "../helpers/onFocusLost";
import onFocusGot from "../helpers/onFocusGot";
import { stringPrettify } from "../helpers/string";
import WUPPopupElement from "../popup/popupElement";
import { WUPcssIcon } from "../styles";
export var SetValueReasons;
(function (SetValueReasons) {
    /** When `control.$value = 'some value'` */
    SetValueReasons[SetValueReasons["manual"] = 1] = "manual";
    /** When clearing happened (by Esc or ClearButton click) OR reset to previous value */
    SetValueReasons[SetValueReasons["clear"] = 2] = "clear";
    /** When user changes on UI */
    SetValueReasons[SetValueReasons["userInput"] = 3] = "userInput";
    /** When user selected existed option on UI and don't need to call selectMenuItem again (for combobox) */
    SetValueReasons[SetValueReasons["userSelect"] = 4] = "userSelect";
    /** When $initValue is changed */
    SetValueReasons[SetValueReasons["initValue"] = 5] = "initValue";
    /** When value changed from storage (on init if `$options.storageKey` is pointed) */
    SetValueReasons[SetValueReasons["storage"] = 6] = "storage";
})(SetValueReasons || (SetValueReasons = {}));
/** Cases of validation for WUP Controls */
export var ValidationCases;
(function (ValidationCases) {
    ValidationCases[ValidationCases["none"] = 0] = "none";
    /** Wait for first user-change > wait for valid > wait for invalid >> show error;
     *  When invalid: wait for valid > hide error > wait for invalid > show error
     *  Also you can check $options.validateDebounceMs */
    ValidationCases[ValidationCases["onChangeSmart"] = 1] = "onChangeSmart";
    /** Validate when user changed value (via type,select etc.); Also you can check $options.validateDebounceMs */
    ValidationCases[ValidationCases["onChange"] = 2] = "onChange";
    /** Validate when control lost focus */
    ValidationCases[ValidationCases["onFocusLost"] = 4] = "onFocusLost";
    /** Validate if control has value and gets focus (recommended option for password with $options.validationShowAll) */
    ValidationCases[ValidationCases["onFocusWithValue"] = 8] = "onFocusWithValue";
    /** Validate when got state 'isReady' and initValue is not empty */
    ValidationCases[ValidationCases["onInit"] = 16] = "onInit";
})(ValidationCases || (ValidationCases = {}));
/** Actions when user pressed ESC or button-clear */
export var ClearActions;
(function (ClearActions) {
    /** Only clear without extra logic */
    ClearActions[ClearActions["clear"] = 0] = "clear";
    /** Pressing Esc/buttonClear: 1st: clear, 2nd: revert clearing (it helps to avoid accidental action) */
    ClearActions[ClearActions["clearBack"] = 1] = "clearBack";
    /** Pressing Esc/buttonClear: 1st: rollback to init, 2nd: clear, 3rd: revert clearing (it helps to avoid accidental action) */
    ClearActions[ClearActions["initClearBack"] = 2] = "initClearBack";
})(ClearActions || (ClearActions = {}));
/** Points on what called validation */
export var ValidateFromCases;
(function (ValidateFromCases) {
    /** When element appended to layout */
    ValidateFromCases[ValidateFromCases["onInit"] = 0] = "onInit";
    /** When control gets focus */
    ValidateFromCases[ValidateFromCases["onFocus"] = 1] = "onFocus";
    /** When control loses focus (including document.activeElement) */
    ValidateFromCases[ValidateFromCases["onFocusLost"] = 2] = "onFocusLost";
    /** When value changed */
    ValidateFromCases[ValidateFromCases["onChange"] = 3] = "onChange";
    /** When form.submit is called (via button submit or somehow else); It's impossible to disable */
    ValidateFromCases[ValidateFromCases["onSubmit"] = 4] = "onSubmit";
    /** When $validate() is called programmatically */
    ValidateFromCases[ValidateFromCases["onManualCall"] = 5] = "onManualCall";
})(ValidateFromCases || (ValidateFromCases = {}));
/** Base abstract form-control */
export default class WUPBaseControl extends WUPBaseElement {
    /** Returns this.constructor // watch-fix: https://github.com/Microsoft/TypeScript/issues/3841#issuecomment-337560146 */
    #ctr = this.constructor;
    /** Text announced by screen-readers when control cleared; @defaultValue `cleared` */
    static $ariaCleared = __wupln("cleared", "aria");
    /** Text announced by screen-readers; @defaultValue `Error for` */
    static $ariaError = __wupln("Error for", "aria");
    /** Css-variables related to component */
    static get $styleRoot() {
        return `:root {
        --ctrl-padding: 1.5em 1em 0.5em 1em;
        --ctrl-focus: var(--base-focus);
        --ctrl-focus-label: var(--ctrl-focus);
        --ctrl-selected: var(--ctrl-focus);
        --ctrl-label: #5e5e5e;
        --ctrl-icon: var(--ctrl-label);
        --ctrl-icon-size: 1em;
        --ctrl-text: inherit;
        --ctrl-bg: #fff;
        --ctrl-border: #e6e6e6;
        --ctrl-border-radius: var(--border-radius);
        --ctrl-err: #ad0000;
        --ctrl-err-bg: #fff4fa;
        --ctrl-err-valid: green;
        --ctrl-invalid-border: red;
        --wup-icon-cross: url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' width='768' height='768'%3E%3Cpath d='M674.515 93.949a45.925 45.925 0 0 0-65.022 0L384.001 318.981 158.509 93.487a45.928 45.928 0 0 0-65.022 0c-17.984 17.984-17.984 47.034 0 65.018l225.492 225.494L93.487 609.491c-17.984 17.984-17.984 47.034 0 65.018s47.034 17.984 65.018 0l225.492-225.492 225.492 225.492c17.984 17.984 47.034 17.984 65.018 0s17.984-47.034 0-65.018L449.015 383.999l225.492-225.494c17.521-17.521 17.521-47.034 0-64.559z'/%3E%3C/svg%3E");
        --wup-icon-check: url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' width='768' height='768'%3E%3Cpath d='M37.691 450.599 224.76 635.864c21.528 21.32 56.11 21.425 77.478 0l428.035-426.23c21.47-21.38 21.425-56.11 0-77.478s-56.11-21.425-77.478 0L263.5 519.647 115.168 373.12c-21.555-21.293-56.108-21.425-77.478 0s-21.425 56.108 0 77.478z'/%3E%3C/svg%3E");
        --wup-icon-dot: url("data:image/svg+xml,%3Csvg viewBox='0 0 100 100' xmlns='http://www.w3.org/2000/svg'%3E%3Ccircle cx='50' cy='50' r='20'/%3E%3C/svg%3E");
        --wup-icon-back: url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' width='768' height='768'%3E%3Cpath d='m509.8 16.068-329.14 329.14c-21.449 21.449-21.449 56.174 0 77.567l329.14 329.14c21.449 21.449 56.174 21.449 77.567 0s21.449-56.174 0-77.567l-290.36-290.36 290.36-290.36c21.449-21.449 21.449-56.173 0-77.567-21.449-21.394-56.173-21.449-77.567 0z'/%3E%3C/svg%3E");
      }
      [wupdark] {
        --ctrl-bg: none;
        --ctrl-border: #104652;
        --ctrl-label: #919191;
        --ctrl-icon: var(--ctrl-label);
        --ctrl-focus-label: #25a1b6;
        --ctrl-selected: #25a1b6;
        --ctrl-err: #ff6666;
        --ctrl-err-bg: #000;
        --ctrl-err-valid: #54d754;
      }`;
        // NiceToHave: change icons to fonts: wupfont: in this case possible to improve btnClear: hover
    }
    /** StyleContent related to component */
    static get $style() {
        // WARN: 'contain:style' is tricky rule
        return `${super.$style}
      :host {
        --ctrl-focus-border: var(--ctrl-focus);
        contain: style;
        display: block;
        flex: 1;
        margin: var(--base-margin) 0;
        box-shadow: 0 0 0 1px var(--ctrl-border);
        border-radius: var(--ctrl-border-radius);
        color: var(--ctrl-text);
        background: var(--ctrl-bg);
        cursor: pointer;
        -webkit-tap-highlight-color: transparent;${
        /* issue: https://stackoverflow.com/questions/25704650/disable-blue-highlight-when-touch-press-object-with-cursorpointer */ ""}
      }
      :host strong,
      :host legend {
        color: var(--ctrl-label);
        pointer-events: none;
      }
      :host[invalid],
      :host[invalid] > [menu] {
        --ctrl-focus-border: var( --ctrl-invalid-border);
        box-shadow: 0 0 3px 0 var(--ctrl-focus-border);
      }
      :host:focus-within,
      :host:focus-within > [menu] {
        box-shadow: 0 0 2px 1px var(--ctrl-focus-border);
      }
      :host:focus-within strong,
      :host:focus-within legend {
        color: var(--ctrl-focus-label);
        pointer-events: initial;
      }
      [disabled] :host,
      :host[disabled] {
        opacity: 0.6;
        cursor: not-allowed;
        -webkit-user-select: none;
        user-select: none;
      }
      [disabled] :host > label,
      :host[disabled] > label {
        pointer-events: none;
      }
      :host label {
        display: flex;
        align-items: center;
        box-sizing: border-box;
        font: inherit;
        cursor: inherit;
        padding: var(--ctrl-padding);
        padding-top: 0;
        padding-bottom: 0;
      }
      :host input,
      :host textarea {
        padding: 0;
        margin: 0;
        cursor: inherit;
      }
      :host [contenteditable=true] {
        margin: var(--ctrl-padding);
        margin-left: 0;
        margin-right: 0;
        cursor: inherit;
      }
      :host strong {
        cursor: inherit;
      }
      :host strong:empty {
        display: none;
      }
      :host [aria-required="true"] + strong:after,
      :host fieldset[aria-required="true"] > legend:after {
        content: "*";
        font-size: larger;
        font-weight: bolder;
        line-height: 0;
      }
      :host [error] {
        cursor: pointer;
        font-size: small;
        color: var(--ctrl-err);
        background: var(--ctrl-err-bg);
        overflow: auto;
        overflow: overlay;
        // text-shadow: 0 0 0 var(--ctrl-err);
      }
      :host [error] ul {
        margin:0; padding:2px 4px 2px;
      }
      :host [error] li {
        display: flex;
        align-items: center;
        margin-left: -5px;
      }
      :host [error] li:before {
        content: '';
        --ctrl-icon-img: var(--wup-icon-dot);
        --ctrl-icon: var(--ctrl-err);
        ${WUPcssIcon}
      }
      :host [error] li[valid] {
        color: var(--ctrl-err-valid);
      }
      :host [error] li[valid]:before {
        content: '';
        --ctrl-icon-img: var(--wup-icon-check);
        --ctrl-icon: var(--ctrl-err-valid);
      }
      :host:focus-within [error] {
        max-height: none;
      }
      @media (hover: hover) and (pointer: fine) {
        :host:hover strong,
        :host[hovered] strong{
          color: var(--ctrl-focus-label);
        }
        :host:hover,
        :host:hover>[menu],
        :host[hovered],
        :host[hovered]>[menu] {
          box-shadow: 0 0 2px 1px var(--ctrl-focus-border);
        }
      }`;
    }
    /** Default function to compare values/changes; It compares by valueOf() & by {id}
     *  Redefine/define `valueOf()` for complex values to improve comparison */
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    static $isEqual(v1, v2, control) {
        return isEqual(v1, v2) || (v1 != null && v1.id != null && v2 != null && v1.id === v2.id);
    }
    /** Provide logic to check if control is empty (by comparison with value) */
    static $isEmpty(v) {
        return v === "" || v === undefined;
    }
    static get observedAttributes() {
        const a = super.observedAttributes;
        a.push("w-initvalue", "disabled", "readonly"); // support for `readonly` & `w-readonly`
        return a;
    }
    static get mappedAttributes() {
        const m = super.mappedAttributes;
        m.autocomplete.type = 2 /* AttributeTypes.string */;
        m.label.type = 2 /* AttributeTypes.string */;
        m.name.type = 2 /* AttributeTypes.string */;
        m.initvalue = {
            type: 5 /* AttributeTypes.parseCustom */,
            parse: (v) => v,
        };
        return m;
    }
    // todo changing global-common-defaults from another project doesn't affect on controls - need to figure out way when user can setup everything in one place
    static $defaults = {
        autoComplete: null,
        autoFocus: false,
        clearActions: 2 /* ClearActions.initClearBack */,
        focusDebounceMs: 100,
        validateDebounceMs: 500,
        validationCase: 1 /* ValidationCases.onChangeSmart */ | 4 /* ValidationCases.onFocusLost */ | 8 /* ValidationCases.onFocusWithValue */,
        validationRules: {
            required: (v, setV) => setV === true && this.$isEmpty(v) && __wupln("This field is required", "validation"),
        },
        validations: null,
        validationShowAll: false,
        disabled: false,
        readOnly: false,
        label: null,
        name: null,
        storage: "local",
        storageKey: "",
    };
    static cloneDefaults() {
        const d = super.cloneDefaults();
        // @ts-expect-error
        d.validationRules = undefined;
        return d;
    }
    /** Called on value change */
    $onChange;
    #value;
    /** Current value of control; You can change it without affecting on $isDirty state */
    get $value() {
        return this.#value;
    }
    set $value(v) {
        this.setValue(v, 1 /* SetValueReasons.manual */);
    }
    #initValue;
    /** Default/init value; used to define isChanged & to reset by keyEsc/buttonClear;
     *  If control not $isDirty and not changed $value is updated according to $initValue */
    get $initValue() {
        return this.#initValue;
    }
    set $initValue(v) {
        const was = this.#initValue;
        const canUpdate = (!this.$isReady && this.$value === undefined) || (!this.$isDirty && !this.$isChanged);
        this.#initValue = v;
        let needUpdate = !this.#ctr.$isEqual(v, was, this); // WARN: comparing required for SelectControl when during the parse it waits for promise
        if (canUpdate && needUpdate) {
            needUpdate = !this.setValue(v, 5 /* SetValueReasons.initValue */);
        }
        needUpdate && this.setClearState();
        if (!this._noDelInitValueAttr) {
            this._isStopChanges = true;
            this.removeAttribute("w-initvalue");
            this._isStopChanges = false;
        }
    }
    #isDirty = false;
    /** True if control is touched by user */
    get $isDirty() {
        return this.#isDirty;
    }
    set $isDirty(v) {
        this.#isDirty = v;
    }
    /** Returns true if value is empty string or undefined */
    get $isEmpty() {
        return this.#ctr.$isEmpty(this.#value);
    }
    /** Returns if value changed (by comparisson with $initValue via static.isEqual option)
     *  By default values compared by valueOf if it's possible */
    get $isChanged() {
        return !this.#ctr.$isEqual(this.$value, this.#initValue, this);
    }
    _isValid;
    /** Returns true if control is valid */
    get $isValid() {
        if (this._isValid == null) {
            this.goValidate(0 /* ValidateFromCases.onInit */, true);
        }
        return this._isValid;
    }
    /** Returns if related form or control disabled (true even if form.$options.disabled && !control.$options.disabled) */
    get $isDisabled() {
        // @ts-expect-error
        const o = this.$form?._opts;
        return o?.disabled || this._opts.disabled || false;
    }
    /** Returns if related form or control readonly (true even if form.$options.readOnly && !control.$options.readOnly) */
    get $isReadOnly() {
        // @ts-expect-error
        const o = this.$form?._opts;
        return o?.readOnly || this._opts.readOnly || false;
    }
    /** Returns if value is required - can't be undefined (depends on $options.validations.required) */
    get $isRequired() {
        return !!this.validations?.required;
    }
    /** Returns autoComplete name if related form or control option is enabled (and control.$options.autoComplete !== false ) */
    get $autoComplete() {
        // @ts-expect-error
        const o = this.$form?._opts;
        const af = this._opts.autoComplete ?? (o?.autoComplete || false);
        return (af === true ? this._opts.name : af) || false;
    }
    /** Check validity and show error if silent is false (by default)
     * @returns errorMessage or false (if valid) */
    $validate(silent = false) {
        return this.goValidate(5 /* ValidateFromCases.onManualCall */, silent);
    }
    /** Check validity and show error
     * @returns errorMessage or false (if valid) */
    validateBySubmit() {
        return this.goValidate(4 /* ValidateFromCases.onSubmit */);
    }
    $showError(err) {
        return this.goShowError(err, this.$refInput);
    }
    $hideError() {
        return this.goHideError();
    }
    #refDetails;
    /** Add (replace) description of control to be announced by screen-readers */
    $ariaDetails(text) {
        // this.setAttr.call(this.$refInput, "aria-description", text); // watchfix: Safari VoiceOver doesn't support aria-description: https://a11ysupport.io/tests/tech__aria__aria-description
        let el = this.#refDetails;
        if (!text) {
            if (el) {
                el.remove();
                this.#refDetails = undefined;
            }
        }
        else {
            if (!el) {
                el = document.createElement("span");
                el.className = this.#ctr.classNameHidden;
                this.#refDetails = this.$refTitle.parentElement.appendChild(el);
            }
            el.textContent = text;
        }
    }
    /** Announce text by screenReaders if element is focused */
    $ariaSpeak(text, delayMs = 100) {
        // don't use speechSynthesis because it's announce despite on screen-reader settings - can be disabled
        // text && speechSynthesis && speechSynthesis.speak(new SpeechSynthesisUtterance(text)); // watchfix: https://stackoverflow.com/questions/72907960/web-accessibility-window-speechsynthesis-vs-role-alert
        const el = document.createElement("section");
        el.setAttribute("aria-live", "off");
        el.setAttribute("aria-atomic", true);
        el.className = this.#ctr.classNameHidden;
        el.id = this.#ctr.$uniqueId;
        const i = this.$refInput;
        const an = "aria-describedby";
        i.setAttribute(an, `${i.getAttribute(an) || ""} ${el.id}`.trimStart());
        this.appendChild(el);
        setTimeout(() => (el.textContent = text), delayMs); // otherwise reader doesn't announce section
        setTimeout(() => {
            el.remove();
            const a = i.getAttribute(an);
            if (a != null) {
                const aNext = a.replace(el.id, "").replace("  ", " ").trim();
                aNext ? i.setAttribute(an, aNext) : i.removeAttribute(an);
            }
        }, 500); // WARN: expected that 500ms is enough to get action from screen-reader
    }
    $form;
    /** Reference to nested HTMLElement */
    $refLabel = document.createElement("label");
    /** Reference to nested HTMLElement */
    $refInput = document.createElement("input");
    /** Reference to nested HTMLElement tied with $options.label */
    $refTitle = document.createElement("strong");
    /** Reference to nested HTMLElement tied with errorMessage */
    $refError;
    // constructor() {
    //   super();
    // }
    gotChanges(propsChanged) {
        super.gotChanges(propsChanged);
        const i = this.$refInput;
        // set label
        const label = (this._opts.label ?? (this._opts.name && __wupln(stringPrettify(this._opts.name), "content"))) || null;
        this.$refTitle.textContent = label;
        const n = !label && this._opts.name ? __wupln(stringPrettify(this._opts.name), "aria") : null;
        this.setAttr.call(i, "aria-label", n);
        // set other props
        this.setAttr("disabled", this._opts.disabled, true);
        this.setAttr("readonly", this._opts.readOnly, true);
        const isReq = this.$isRequired;
        this.setAttr("required", isReq, true);
        this.setAttr.call(this.$refInput, "aria-required", isReq);
        this.setupInitValue(propsChanged);
        propsChanged?.includes("clearActions") && this.setClearState();
        this.gotFormChanges(propsChanged);
    }
    /** Called on control/form Init and every time as control/form options changed. Method contains changes related to form `disabled`,`readonly` etc. */
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    gotFormChanges(propsChanged) {
        this.setupInput();
        this.$refError && !this.canShowError && this.goHideError(); // hide error if user can't touch the control
    }
    /** Called to update disabled/readonly/autocomplete options on input */
    setupInput() {
        const i = this.$refInput;
        i.disabled = this.$isDisabled;
        // @ts-ignore
        i.autocomplete = this.$autoComplete || "off";
        this.setupInputReadonly();
    }
    /** Called to update readonly option on input */
    setupInputReadonly() {
        this.$refInput.readOnly = this.$isReadOnly;
    }
    /** Called on Init and options/attributes changes to update $initValue or $value (if pointed storageKey) */
    setupInitValue(propsChanged) {
        // lowercase for attribute-changes otherwise it's wrong
        if (!propsChanged || propsChanged.includes("initvalue")) {
            // @ts-expect-error - because initvalue not defined but possible from attributes
            const attr = this._opts.initvalue; // this value from attribute in string
            if (attr != null) {
                this._noDelInitValueAttr = true;
                try {
                    this.$initValue = this.parse(attr);
                }
                catch (err) {
                    this.throwError(err);
                }
                delete this._noDelInitValueAttr;
            }
            else if (propsChanged) {
                this.$initValue = undefined; // removed attr >> remove initValue
            }
        }
        // retrieve value from model
        if (this.$initValue === undefined && this.$form && this._opts.name && !this.hasAttribute("initvalue")) {
            if (!propsChanged || propsChanged.includes("name")) {
                this.$initValue = nestedProperty.get(this.$form._initModel, this._opts.name);
            }
        }
        // retrieve value from store
        if (!propsChanged && this._opts.storageKey) {
            this.setValue(this.storageGet(), 6 /* SetValueReasons.storage */);
        }
    }
    /** Returns true on !$isDisabled */
    get canShowError() {
        return !this.$isDisabled; // && !this.$isReadOnly;
    }
    gotReady() {
        super.gotReady();
        const onFocusGotHandler = (e) => {
            const arr = this.gotFocus(e);
            const r = onFocusLostEv(this, () => {
                this.gotFocusLost();
                arr.forEach((f) => f());
                r();
                this.disposeLst.splice(this.disposeLst.indexOf(r), 1);
            });
            this.disposeLst.push(r);
        };
        this.$isFocused && setTimeout(onFocusGotHandler); // case if append element and focus immediately
        const r = onFocusGot(this, onFocusGotHandler, { debounceMs: this._opts.focusDebounceMs });
        this.disposeLst.push(r);
        // appendEvent removed by dispose()
        this.appendEvent(this, "mousedown", // to prevent blur-focus effect for input by label click
        (e) => {
            if (this.$isDisabled) {
                return;
            }
            const el = e.target;
            !(el.tagName === "INPUT" ||
                el.tagName === "TEXTAREA" ||
                this.$refInput === el ||
                this.includes.call(this.$refInput, el)) && e.preventDefault();
            // this required because default focus-effect was prevented
            this.appendEvent(this, "mouseup", () => !this.$isFocused && this.focus(), { once: true });
        }, { passive: false } // otherwise preventDefault doesn't work
        );
        if (this._opts.validationCase & 16 /* ValidationCases.onInit */) {
            !this.$isEmpty && this.goValidate(0 /* ValidateFromCases.onInit */);
        }
    }
    gotRender() {
        this.renderControl();
    }
    connectedCallback() {
        delete this._errName;
        delete this._errMsg;
        delete this._wasValidNotEmpty;
        super.connectedCallback();
        this.$form = WUPFormElement.$tryConnect(this);
        this.#isDirty = false; // reset state after re-appended
    }
    gotRemoved() {
        super.gotRemoved();
        this.$form?.$controls.splice(this.$form.$controls.indexOf(this), 1);
    }
    /** Returns validations enabled by user & defaults */
    get validations() {
        const def = this.#ctr.$defaults.validations;
        const opt = this._opts.validations;
        return def ? { ...def, ...opt } : opt;
    }
    /** Returns validations functions ready for checking */
    get validationsRules() {
        const vls = this.validations;
        if (!vls) {
            return [];
        }
        const rules = this.#ctr.$defaults.validationRules;
        const check = (v, k, reason) => {
            const vl = vls[k];
            let err;
            if (typeof vl === "function") {
                err = vl(v, this, reason);
            }
            else {
                const r = rules[k];
                if (!r) {
                    const n = this._opts.name ? `.[${this._opts.name}]` : "";
                    throw new Error(`${this.tagName}${n}. Validation rule [${k}] is not found`);
                }
                err = r.call(this, v, vl, this, reason);
            }
            if (err !== false) {
                return err;
            }
            return false;
        };
        // eslint-disable-next-line @typescript-eslint/no-this-alias
        const self = this;
        const arr = Object.keys(vls)
            .sort((k1, k2) => {
            if (k1 === "required")
                return -1;
            if (k2 === "required")
                return 1;
            return 0;
        })
            .map((key) => ({ [key]: (v, reason) => check.call(self, v, key, reason) }[key])); // make object to create named function
        return arr;
    }
    /** Current name of failed validation */
    _errName;
    _wasValidNotEmpty;
    _validTimer;
    /** Method called to check control based on validation rules and current value */
    goValidate(fromCase, silent = false) {
        this._errName = undefined;
        const vls = this.validationsRules;
        if (!vls.length) {
            this._isValid = true;
            if (!this.$isEmpty) {
                this._wasValidNotEmpty = true;
            }
            return false;
        }
        const v = this.$value;
        let errMsg = "";
        const isEmpty = this.$isEmpty;
        this._isValid = !vls.some((fn) => {
            // process empty value only on rule 'required'; for others skip if value is empty
            const skipRule = isEmpty && fn.name[0] !== "_" && fn.name !== "required"; // undefined only for 'required' rule; for others: skip if value = undefined
            const err = !skipRule && fn(v, fromCase);
            if (err) {
                this._errName = fn.name;
                errMsg = err;
                return true;
            }
            return false;
        });
        this._validTimer && clearTimeout(this._validTimer);
        let canShowError = (!silent || this.$refError) && this.canShowError;
        if (fromCase === 3 /* ValidateFromCases.onChange */ && this._opts.validationCase & 1 /* ValidationCases.onChangeSmart */) {
            if (errMsg && !this._wasValidNotEmpty && !this.$refError) {
                canShowError = false;
            }
        }
        if (this._isValid && !isEmpty) {
            this._wasValidNotEmpty = true;
        }
        if (errMsg) {
            if (canShowError) {
                // don't wait for debounce if we need to update an error
                const waitMs = this.$refError || fromCase !== 3 /* ValidateFromCases.onChange */ ? 0 : this._opts.validateDebounceMs;
                const saved = this._errName;
                const act = () => {
                    this.goShowError(errMsg, this.$refInput);
                    this._errName = saved;
                };
                this._validTimer = waitMs ? setTimeout(act, waitMs) : (act(), undefined);
            }
            return errMsg;
        }
        this.goHideError();
        return false;
    }
    /** Show (append/update) all validation-rules with checkpoints to existed error-element */
    renderValidations(parent, skipRules = ["required"]) {
        const vls = this.validationsRules.filter((vl) => !skipRules.includes(vl.name) && vl.name[0] !== "_");
        if (!vls.length) {
            return;
        }
        let p = parent;
        if (!p._wupVldItems) {
            p = parent;
            p._wupVldItems = [];
            const ul = p.appendChild(document.createElement("ul"));
            ul.setAttribute("aria-hidden", true);
            for (let i = 0; i < vls.length; ++i) {
                const li = ul.appendChild(document.createElement("li"));
                li._wupVld = vls[i];
                p._wupVldItems.push(li);
                const err = vls[i](undefined, null);
                if (err) {
                    li.textContent = err;
                }
                else {
                    // eslint-disable-next-line no-loop-func
                    setTimeout(() => {
                        const n = this._opts.name ? `.[${this._opts.name}]` : "";
                        throw new Error(`${this.tagName}${n}. Can't get error message for validationRule [${vls[i].name}]. Calling rule with value [undefined] must return error message`);
                    });
                }
            }
        }
        const v = this.$value;
        p._wupVldItems.forEach((li) => this.setAttr.call(li, "valid", li._wupVld(v, null) === false, true));
    }
    renderError() {
        const p = document.createElement("wup-popup");
        p.$options.openCase = 1 /* PopupOpenCases.onInit */;
        p.$options.target = this;
        p.$options.placement = [
            WUPPopupElement.$placements.$bottom.$start.$resizeWidth,
            WUPPopupElement.$placements.$top.$start.$resizeWidth,
        ];
        p.$options.offset = [-2, 0];
        p.setAttribute("error", "");
        // p.setAttribute("role", "alert");
        p.setAttribute("aria-live", "off");
        p.setAttribute("aria-atomic", true); // necessary to make Voiceover on iOS read the error messages after more than one invalid submission
        p.id = this.#ctr.$uniqueId;
        const hiddenLbl = p.appendChild(document.createElement("span"));
        hiddenLbl.className = this.#ctr.classNameHidden;
        p.appendChild(document.createElement("span"));
        this.$refError = this.appendChild(p);
        return p;
    }
    /** Current error message */
    _errMsg;
    #refErrTarget;
    /** Method called to show error and set invalid state on input; point null to show all validation rules with checkpoints */
    goShowError(err, target) {
        if (!err) {
            this.throwError(new Error("missed error message"));
            return;
        }
        if (this._errMsg === err) {
            return;
        }
        // possible when user goes to another page and focusout > validTimeout happened
        if (!this.isConnected) {
            return;
        }
        this._errName = undefined;
        this._errMsg = err;
        if (!this.$refError) {
            this.$refError = this.renderError();
        }
        this._opts.validationShowAll && this.renderValidations(this.$refError);
        this.#refErrTarget = target;
        target.setCustomValidity?.call(target, err);
        this.setAttribute("invalid", "");
        const lbl = this.$refTitle.textContent;
        this.$refError.firstElementChild.textContent = lbl ? `${this.#ctr.$ariaError} ${lbl}:` : "";
        const el = this.$refError.children.item(1);
        el.textContent = err;
        const renderedErr = this.$refError._wupVldItems?.find((li) => li.textContent === err);
        this.setAttr.call(el, "class", renderedErr ? this.#ctr.classNameHidden : null);
        target.setAttribute("aria-describedby", this.$refError.id); // watchfix: nvda doesn't read aria-errormessage: https://github.com/nvaccess/nvda/issues/8318
        if (!this.$isFocused) {
            this.$refError.setAttribute("role", "alert"); // force to announce error when focus is already missed
            renderedErr?.scrollIntoView();
        }
    }
    /** Method called to hide error and set valid state on input */
    goHideError() {
        this._errMsg = undefined;
        if (this.$refError) {
            const p = this.$refError;
            p.addEventListener("$close", p.remove, { passive: true, once: true });
            p.$close(); // hide with animation
            this.$refError = undefined;
            this.#refErrTarget.setCustomValidity?.call(this.#refErrTarget, "");
            this.#refErrTarget.removeAttribute("aria-describedby");
            this.removeAttribute("invalid");
            this.#refErrTarget = undefined;
        }
    }
    /** Called to serialize value from URL/storage; override it if you have object */
    valueFromUrl(str) {
        return str === "null" ? null : this.parse(str);
    }
    /** Called to serialize value to URL/storage & must return null if need to remove */
    valueToUrl(v) {
        if (v == null) {
            return "null";
        }
        if (typeof v === "object") {
            this.throwError(new Error("Option `storageKey` with object-values are not supported. Override `valueToUrl` & `valueFromUrl` to fix it"));
            return null;
        }
        return v.toString();
    }
    /** Returns storage key based on options `storageKey` and `name` */
    get storageKey() {
        return this._opts.storageKey === true ? this._opts.name : this._opts.storageKey;
    }
    /** Get & parse value from storage according to options `storageKey`, `storage` and `name` */
    storageGet() {
        const key = this.storageKey;
        if (key) {
            let v;
            switch (this._opts.storage) {
                case "session":
                    v = window.sessionStorage.getItem(key);
                    break;
                case "url":
                    v = new URLSearchParams(window.location.search).get(decodeURIComponent(key).replace(/\+/g, " ")); // regex require for form-url-decode 's+t' => 's t'
                    v = v != null ? decodeURIComponent(v) : v;
                    break;
                default:
                    v = window.localStorage.getItem(key);
                    break;
            }
            if (v != null) {
                return this.valueFromUrl(v);
            }
        }
        return this.$initValue;
    }
    /** Save value to storage storage according to options `storageKey`, `storage` and `name` */
    storageSet(v) {
        // NiceToHave: option to sync ctrls with same storage & key: if 'personType` is changed need to update all ctrls with same key `personType`
        const key = this.storageKey;
        if (!key) {
            return; // possible when _opts.name is empty
        }
        try {
            let strg;
            switch (this._opts.storage) {
                case "session":
                    strg = window.sessionStorage;
                    break;
                case "url":
                    {
                        const url = new URL(window.location.href);
                        strg = {
                            removeItem: (k) => {
                                url.searchParams.delete(k);
                                window.history.replaceState(null, "", url); // OR window.history.pushState(null, null, url);
                            },
                            setItem: (k, val) => {
                                url.searchParams.set(k, val);
                                window.history.replaceState(null, "", url); // OR window.history.pushState(null, null, url);
                            },
                        };
                    }
                    break;
                default:
                    strg = window.localStorage;
                    break;
            }
            if (this.#ctr.$isEmpty(v)) {
                strg.removeItem(key);
            }
            else {
                const sv = this.valueToUrl(v);
                sv === null ? strg.removeItem(key) : strg.setItem(key, sv);
            }
        }
        catch (err) {
            this.throwError(err); // re-throw error when storage is full
        }
    }
    /** Fire this method to update value & validate; returns null when not $isReady, true if changed */
    setValue(v, reason) {
        const prev = this.#value;
        this.#value = v;
        if (!this.$isReady) {
            this.setClearState();
            return null;
        }
        if (reason !== 5 /* SetValueReasons.initValue */ &&
            reason !== 1 /* SetValueReasons.manual */ &&
            reason !== 6 /* SetValueReasons.storage */) {
            this.$isDirty = true;
        }
        const isChanged = !this.#ctr.$isEqual(v, prev, this);
        if (!isChanged) {
            return false;
        }
        this.setClearState();
        this._isValid = undefined;
        const canVld = reason !== 1 /* SetValueReasons.manual */;
        (canVld || this.$refError) && this.validateAfterChange();
        if (reason !== 5 /* SetValueReasons.initValue */) {
            // save to storage
            reason !== 6 /* SetValueReasons.storage */ && this._opts.storageKey && this.storageSet(v);
            setTimeout(() => this.fireEvent("$change", { cancelable: false, bubbles: true, detail: { reason } }));
        }
        return true;
    }
    /** Called after value is changed */
    validateAfterChange() {
        const c = this._opts.validationCase;
        if (c & 2 /* ValidationCases.onChange */ || c & 1 /* ValidationCases.onChangeSmart */) {
            this.goValidate(3 /* ValidateFromCases.onChange */);
        }
    }
    _nextClearValue;
    /** Called every time on value-change to update clear-state & buttonClear
     * @returns nextClearValue */
    setClearState() {
        const clr = this._opts.clearActions;
        switch (clr) {
            case 1 /* ClearActions.clearBack */:
                if (!this.$isEmpty) {
                    delete this._nextClearValue;
                }
                break;
            case 2 /* ClearActions.initClearBack */:
                if (this.$isChanged) {
                    this._nextClearValue = this.$initValue;
                }
                else if (!this.$isEmpty) {
                    delete this._nextClearValue;
                }
                break;
            default:
                delete this._nextClearValue;
                break;
        }
        return this._nextClearValue;
    }
    /* Called when user pressed Esc-key or button-clear */
    clearValue() {
        const next = this._nextClearValue;
        this._nextClearValue = this.#value;
        this.setValue(next, 2 /* SetValueReasons.clear */);
        this.$isEmpty ? this.$ariaSpeak(this.#ctr.$ariaCleared) : this.$refInput.select();
    }
    /** Called when element got focus; must return array of RemoveFunctions called on FocusLost */
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    gotFocus(ev) {
        this.$refError?.$refresh();
        const c = this._opts.validationCase;
        if (c & 8 /* ValidationCases.onFocusWithValue */ && !this.$isEmpty) {
            this.goValidate(1 /* ValidateFromCases.onFocus */);
        }
        else if (this._wasValidNotEmpty == null && c & 1 /* ValidationCases.onChangeSmart */) {
            this.goValidate(3 /* ValidateFromCases.onChange */, true); // validate to define current state
        }
        const r = this.appendEvent(this, "keydown", (e) => !this.$isDisabled && !this.$isReadOnly && this.gotKeyDown(e), {
            passive: false,
        });
        return [r];
    }
    /** Called when element completely lost focus; despite on blur it has debounce filter */
    gotFocusLost() {
        if (this._opts.validationCase & 4 /* ValidationCases.onFocusLost */) {
            // const wasErr = !!this._errMsg;
            this.goValidate(2 /* ValidateFromCases.onFocusLost */);
            // NiceToHave test & rollback it when error will be visible over popup: for selectControl the color is changed but error not visible and menu stays opened
            // this._errMsg && !wasErr && this.focus(); // user sees validation error at first time: return focus back once
        }
    }
    /** Called when user pressed key */
    gotKeyDown(e) {
        const canClear = !e.defaultPrevented && e.key === "Escape" && !e.shiftKey && !e.altKey && !e.ctrlKey;
        // WARN: Escape works wrong with NVDA because it enables NVDA-focus-mode
        canClear && setTimeout(() => !e.defaultPrevented && this.clearValue()); // timeout to wait for modal to handle it
    }
}
// NiceToHave when control is disabled need to show tooltip with reason
// NiceToHave when control is readonly need to show tooltip with reason
