import nestedProperty from "../helpers/nestedProperty";
import onEvent from "../helpers/onEvent";
import promiseWait from "../helpers/promiseWait";
import WUPSpinElement from "../spinElement";
import { WUPcssMenu } from "../styles";
import WUPBaseComboControl from "./baseCombo";
const tagName = "wup-select";
/** Form-control with dropdown/combobox behavior
 * @example
  const el = document.createElement("wup-select");
  el.$options.name = "gender";
  el.$options.items = [
    { value: 1, text: "Male" },
    { value: 2, text: "Female" },
    { value: 3, text: "Other/Skip" },
  ];
  el.$initValue = 3;
  el.$options.validations = { required: true };
  const form = document.body.appendChild(document.createElement("wup-form"));
  form.appendChild(el);
  // or HTML
  <wup-form>
    <wup-select w-name="gender" w-initvalue="3" w-validations="myValidations" w-items="myDropdownItems" />
  </wup-form>;
 * @tutorial innerHTML @example
 * <label>
 *   <span> // extra span requires to use with icons via label:before, label:after without adjustments
 *      <input/>
 *      <strong>{$options.label}</strong>
 *   </span>
 *   <button clear/>
 *   <wup-popup menu>
 *      <ul>
 *          <li>Item 1</li>
 *          <li>Item 2</li>
 *          // etc. /
 *      </ul>
 *   </wup-popup>
 * </label>
 */
export default class WUPSelectControl extends WUPBaseComboControl {
    /** Returns this.constructor // watch-fix: https://github.com/Microsoft/TypeScript/issues/3841#issuecomment-337560146 */
    #ctr = this.constructor;
    static get $styleRoot() {
        return `:root {
        --ctrl-select-icon-img: url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' width='768' height='768'%3E%3Cpath d='m16.078 258.214 329.139 329.139c21.449 21.449 56.174 21.449 77.567 0l329.139-329.139c21.449-21.449 21.449-56.174 0-77.567s-56.174-21.449-77.567 0L384 471.003 93.644 180.647c-21.449-21.449-56.173-21.449-77.567 0s-21.449 56.173 0 77.567z'/%3E%3C/svg%3E");
      }`;
    }
    // WARN: scroll sets for ul otherwise animation broken scroll to selectItem because animation affects on scrollSize
    static get $style() {
        return `${super.$style}
      :host {
        --ctrl-icon-img: var(--ctrl-select-icon-img);
      }
      :host label:after {
        height: 100%;
        align-self: center;
      }
      :host[opened] label:after {
        transform: rotate(180deg);
      }
      ${WUPcssMenu(":host [menu]")}`;
    }
    /** Text for listbox when no items are displayed */
    static $textNoItems = __wupln("No Items", "content");
    /** Text for aria-label of <ul> element */
    static $ariaLabelItems = __wupln("Items", "aria");
    static $isEqual(v1, v2, c) {
        let isEq = super.$isEqual(v1, v2, c);
        if (!isEq && c._opts.multiple && Array.isArray(v1) && Array.isArray(v2)) {
            const isSortMatter = c._opts.sortable; // option sortable is defined in selectMany but placed here to simplify logic
            isEq =
                v1.length === v2.length && (isSortMatter ? v1.every((v, i) => v2[i] === v) : v1.every((v) => v2.includes(v)));
        }
        return isEq;
    }
    /** Function to filter menuItems based on inputValue
     * @param menuItemText textcontent in lowercase
     * @param menuItemValue value related to items[x].value
     * @param inputValue normalized input value (trimStart + lowercase)
     * @param inputRawValue input value without normalization
     * @returns true if menuItem must be visible in menu
     */
    static $filterMenuItem(menuItemText, menuItemValue, inputValue, 
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    inputRawValue) {
        return !inputValue || menuItemText.startsWith(inputValue) || menuItemText.includes(` ${inputValue}`);
    }
    static $defaults = {
        ...WUPBaseComboControl.$defaults,
        validationRules: {
            ...WUPBaseComboControl.$defaults.validationRules,
            minCount: (v, setV) => (v == null || v.length < setV) && __wupln(`Min count is ${setV}`, "validation"),
            maxCount: (v, setV) => (v == null || v.length > setV) && __wupln(`Max count is ${setV}`, "validation"),
        },
        openCase: 8 /* MenuOpenCases.onClick */ | 2 /* MenuOpenCases.onFocus */ | 64 /* MenuOpenCases.onPressArrowKey */ | 32 /* MenuOpenCases.onInput */,
        allowNewValue: false,
        multiple: false,
        items: [],
    };
    static cloneDefaults() {
        const d = super.cloneDefaults();
        d.items = [];
        return d;
    }
    /** Returns whether control in pending state or not (shows spinner) */
    get $isPending() {
        return !!this.#stopPending;
    }
    /** Called when need to parse attr [initValue] */
    parse(attrValue) {
        if (this._opts.multiple) {
            return nestedProperty.get(window, attrValue);
        }
        // WARN: parse must be called only after items is fetched
        const arr = this.getItems();
        if (arr?.length) {
            const r = attrValue === ""
                ? arr.find((o) => o.value == null)
                : arr.find((o) => o.value && `${o.value}` === attrValue);
            return r?.value;
        }
        return undefined;
    }
    parseInput(text) {
        let v;
        if (!this._opts.multiple) {
            const s = text.trim();
            v = s ? this.findValueByText(s) ?? s : undefined;
        }
        else {
            // WARN it works only with allowNewValue
            v = text.split(",").map((si) => si.trim());
            v = v
                .filter((si, i) => v.indexOf(si) === i) // allow only unique values
                .map((si) => {
                if (!si) {
                    return undefined;
                }
                const vi = this.findValueByText(si);
                return vi ?? (this._opts.allowNewValue ? si : vi);
            })
                .filter((vi) => vi !== undefined);
            if (!v.length) {
                return undefined;
            }
        }
        return v;
    }
    valueFromUrl(str) {
        if (this.$options.multiple) {
            this._opts.multiple = false;
            const v = str.split("_").map((si) => super.valueFromUrl(si));
            this._opts.multiple = true; // otherwise parse is wrong
            return v;
        }
        return super.valueFromUrl(str);
    }
    valueToUrl(v) {
        if (this.$options.multiple) {
            return v.map((vi) => super.valueToUrl(vi)).join("_");
        }
        return super.valueToUrl(v);
    }
    /** It's called with option allowNewValue to find value related to text */
    findValueByText(txt) {
        const s = txt.toLowerCase();
        const i = this._menuItems?.all.findIndex((o) => o._text === s);
        const r = i > -1 ? this._cachedItems[i].value : undefined;
        return r;
    }
    /** Called on every spin-render */
    renderSpin() {
        WUPSpinElement.$use();
        const spin = document.createElement("wup-spin");
        spin.$options.fit = true;
        spin.$options.overflowFade = true;
        spin.$options.overflowTarget = this;
        this.appendChild(spin);
        return spin;
    }
    /** Required to wait for fetching items to setup initValue */
    _onPendingInitValue;
    /** Called to get/fetch items based on $options.items */
    fetchItems() {
        this._cachedItems = undefined;
        this.#findValT && clearTimeout(this.#findValT); // prevent setInputValue if fetchItems is started
        this.#findValT = undefined;
        const { items } = this._opts;
        let d = items;
        if (typeof d === "function") {
            d = d();
        }
        const act = () => {
            this._onPendingInitValue?.call(this);
            delete this._onPendingInitValue;
            this.setInputValue(this.$value, 5 /* SetValueReasons.initValue */);
            this.setupInputReadonly(); // call it because opt readonlyInput can depend on items.length
        };
        if (d instanceof Promise) {
            return promiseWait(d, 300, (v) => this.changePending(v))
                .then((data) => {
                this._cachedItems = data || [];
                return data;
            })
                .finally(() => {
                this._cachedItems ??= [];
                act();
                this.$isFocused && this.goOpenMenu(2 /* MenuOpenCases.onFocus */, null);
            });
        }
        this._cachedItems = d;
        act();
        return d;
    }
    gotChanges(propsChanged) {
        this._opts.items ??= [];
        if (!propsChanged || propsChanged.includes("items")) {
            this.removePopup();
            this.fetchItems(); // WARN: it's important to be before super otherwise initValue won't work
        }
        super.gotChanges(propsChanged);
    }
    setupInputReadonly() {
        const r = this._opts.readOnlyInput;
        this.$refInput.readOnly =
            this.$isReadOnly ||
                this.$isPending ||
                r === true ||
                (typeof r === "number" && r > (this._cachedItems?.length || 0) && !this._opts.allowNewValue); // WARN: _cached items can be undefined when fetching not started yet
    }
    setupInitValue(propsChanged) {
        if (this._cachedItems) {
            super.setupInitValue(propsChanged);
        }
        else {
            // case1: attr [initValue] > parse > setValue
            // case2: $value > setValue
            this._onPendingInitValue = () => super.setupInitValue(propsChanged);
        }
    }
    #stopPending;
    /** Change pending state */
    changePending(v) {
        if (v === !!this.#stopPending) {
            return;
        }
        if (v) {
            this.$refInput.setAttribute("aria-busy", true);
            const refSpin = this.renderSpin();
            this.$refInput.readOnly = true;
            this.#stopPending = () => {
                this.$refInput.readOnly = this.$isReadOnly;
                this.#stopPending = undefined;
                this.$refInput.removeAttribute("aria-busy");
                refSpin.remove();
            };
        }
        else {
            this.#stopPending();
        }
    }
    /** All items of current menu */
    _menuItems;
    /** Items resolved from options */
    _cachedItems;
    /** Called when NoItems need to show */
    renderMenuNoItems(popup, isReset) {
        if (isReset) {
            popup.hidden = false;
            const liSaved = this._menuItems._refNoItems;
            if (liSaved) {
                liSaved.style.display = "none";
            }
        }
        else if (this.#ctr.$textNoItems) {
            const liSaved = this._menuItems._refNoItems;
            if (liSaved) {
                liSaved.style.display = "";
            }
            else {
                const ul = popup.children.item(0);
                const li = ul.appendChild(document.createElement("li"));
                li.textContent = this.#ctr.$textNoItems;
                li.setAttribute("role", "option");
                li.setAttribute("aria-disabled", true);
                li.setAttribute("aria-selected", false);
                this._menuItems._refNoItems = li;
            }
        }
        else {
            popup.hidden = true;
        }
    }
    renderMenu(popup, menuId) {
        const ul = popup.appendChild(document.createElement("ul"));
        ul.setAttribute("id", menuId);
        ul.setAttribute("role", "listbox");
        ul.setAttribute("aria-label", this.#ctr.$ariaLabelItems);
        ul.setAttribute("tabindex", -1); // otherwise Firefox move focus into it by keydown 'tab'
        this.setAttr.call(ul, "aria-multiselectable", this._opts.multiple === true);
        const all = this.renderMenuItems(ul);
        this._menuItems = { all, focused: -1 };
        !all.length && this.renderMenuNoItems(popup, false);
        // it happens on every show because by hide it dispose events
        onEvent(ul, "click", (e) => {
            e.preventDefault(); // to prevent popup-hide-show
            let t = e.target;
            if (t === ul) {
                return;
            }
            while (1) {
                const parent = t.parentElement;
                if (parent === ul) {
                    const isDisabled = t.hasAttribute("aria-disabled");
                    !isDisabled && this.gotMenuItemClick(e, t);
                    break;
                }
                t = parent;
            }
        }, { passive: false });
        return ul;
    }
    /** Method returns items defined based on $options.items */
    getItems() {
        if (!this._cachedItems) {
            this.throwError(new Error("Internal bug. No cached items"));
            return [];
        }
        return this._cachedItems;
    }
    /* Called when need to show text related to value */
    valueToText(v, items) {
        const i = items.findIndex((o) => this.#ctr.$isEqual(o.value, v, this));
        if (i === -1) {
            if (this._opts.allowNewValue) {
                return v != null ? v.toString() : "";
            }
            console.error(`${this.tagName}${this._opts.name ? `[${this._opts.name}]` : ""}. Not found in items`, {
                items,
                value: v,
            });
            return `Error: not found for ${v != null ? v.toString() : ""}`;
        }
        const item = items[i];
        if (typeof item.text === "function") {
            const li = document.createElement("li");
            const s = item.text(item.value, li, i);
            li.remove();
            return s;
        }
        return item.text;
    }
    #findValT; // need to clear previous to avoid collission
    setInputValue(v, reason) {
        this.#findValT && clearTimeout(this.#findValT);
        this.#findValT = undefined;
        if (this.$isPending) {
            return; // decline because setInputValue will be fired after pending
        }
        if (reason === 1 /* SetValueReasons.manual */ || reason === 5 /* SetValueReasons.initValue */) {
            this.#findValT = setTimeout(() => super.setInputValue(v, reason), 2); // timeout to fix case when el.$options.items=... el.$value=...
        }
        else {
            super.setInputValue(v, reason);
        }
    }
    valueToInput(v) {
        if (v === undefined || (this._opts.multiple && !v.length)) {
            return "";
        }
        // WARN: possible case when value changed but prev getItems.then still in progress
        const items = this.getItems();
        if (this._opts.multiple) {
            const s = v?.map((vi) => this.valueToText(vi, items)).join(", ");
            return this.$isFocused ? `${s}, ` : s;
        }
        return this.valueToText(v, items);
    }
    /** Create menuItems as array of HTMLLiElement with option _text required to filtering by input (otherwise content can be html-structure) */
    renderMenuItems(ul) {
        const arr = this.getItems();
        const arrLi = arr.map((a) => {
            const li = ul.appendChild(document.createElement("li"));
            li.setAttribute("role", "option");
            // li.setAttribute("aria-selected", "false");
            li._value = a.value;
            return li;
        });
        if (arr.length) {
            if (typeof arr[0].text === "function") {
                arr.forEach((v, i) => {
                    arrLi[i]._text = v.text(v.value, arrLi[i], i).toLowerCase();
                });
            }
            else {
                arr.forEach((v, i) => {
                    const txt = v.text;
                    arrLi[i].textContent = txt;
                    arrLi[i]._text = txt.toLowerCase();
                });
            }
        }
        return arrLi;
    }
    /** Called when need to setValue & close base on clicked item */
    gotMenuItemClick(_e, item) {
        const i = this._menuItems.all.indexOf(item);
        const o = this._cachedItems[i];
        const canOff = this._opts.multiple && item.getAttribute("aria-selected") === "true";
        this.selectMenuItem(canOff ? null : item); // select/deselect
        canOff && item.setAttribute("aria-selected", "false");
        this.selectValue(o.value, !this._opts.multiple);
    }
    selectValue(v, canCloseMenu = true) {
        if (this._opts.multiple) {
            let arr = this.$value;
            if (!arr?.length) {
                v = [v];
            }
            else {
                const i = arr.indexOf(v);
                if (i !== -1) {
                    if (arr.length === 1) {
                        arr = undefined;
                        // this.selectMenuItem(null);
                    }
                    else {
                        arr = [...arr];
                        arr.splice(i, 1);
                    }
                }
                else {
                    arr = [...arr, v];
                }
                v = arr;
            }
        }
        super.selectValue(v, canCloseMenu);
        this._opts.multiple && this.clearFilterMenuItems();
    }
    canOpenMenu(openCase, e) {
        if (this.$isPending) {
            return false;
        }
        return super.canOpenMenu(openCase, e);
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    async goOpenMenu(openCase, e) {
        if (this.$isOpened) {
            return this.$refPopup;
        }
        if (this.$refPopup && openCase !== 32 /* MenuOpenCases.onInput */ && this._menuItems.filtered) {
            this.clearFilterMenuItems();
        }
        const popup = await super.goOpenMenu(openCase, e);
        popup && !this.canClearSelection && this.selectMenuItemByValue(this.$value); // set aria-selected only if input not empty
        return popup;
    }
    /** Returns first selected visible item */
    getFirstSelected() {
        if (this._selectedMenuItem) {
            const items = this._selectedMenuItem.parentElement.querySelectorAll("[aria-selected=true]");
            for (let i = 0, item = items[0]; i < items.length; item = items[++i]) {
                if (!item.style.display) {
                    return item;
                }
            }
        }
        return undefined;
    }
    selectMenuItem(next) {
        if (this._opts.multiple) {
            this._selectedMenuItem = undefined; // otherwise multiple selection not allowed in combobox
        }
        super.selectMenuItem(next);
    }
    /** Select menu item if value !== undefined and menu is opened */
    selectMenuItemByValue(v) {
        if (this.$isOpened) {
            const findAndSelect = (vi) => {
                const i = this._cachedItems.findIndex((item) => this.#ctr.$isEqual(item.value, vi, this));
                this.selectMenuItem(this._menuItems.all[i] || null);
                return i;
            };
            v === undefined && this._selectedMenuItem && this.selectMenuItem(null);
            if (!this._opts.multiple) {
                v !== undefined && findAndSelect(v);
            }
            else {
                const set = new Set();
                v !== undefined && v.forEach((vi) => set.add(findAndSelect(vi)));
                // deselect others
                this._menuItems.all.forEach((el, i) => !set.has(i) && el.hasAttribute("aria-selected") && el.removeAttribute("aria-selected"));
            }
        }
    }
    focusMenuItem(next) {
        next && !next.hasAttribute("aria-selected") && next.setAttribute("aria-selected", false);
        super.focusMenuItem(next);
        if (next === null && this._menuItems) {
            this._menuItems.focused = -1;
        }
    }
    /** Focus item by index or reset is index is null (via aria-activedescendant).
     *  If menuItems is filtered by input-text than index must point on filtered array */
    focusMenuItemByIndex(index) {
        const { filtered } = this._menuItems;
        const trueIndex = filtered ? filtered[index] : index;
        const next = this._menuItems.all[trueIndex];
        this.focusMenuItem(next);
        this._menuItems.focused = index;
    }
    focusMenuItemByKeydown(e) {
        if (!this._menuItems) {
            return; // possible when user goes to another but menuItems still is loading
            // throw new Error(`${this.tagName}. Impossible on empty menu`);
        }
        const { length } = this._menuItems.filtered || this._menuItems.all;
        let focusIndex = null;
        // firstFocused can be selected/current
        if (this._menuItems.focused === -1) {
            if (e.key === "ArrowDown" || e.key === "ArrowUp") {
                const item = this.getFirstSelected();
                if (item) {
                    focusIndex = this._menuItems.all.indexOf(item);
                }
            }
        }
        if (focusIndex === null) {
            switch (e.key) {
                case "ArrowDown":
                    {
                        let cur = this._menuItems.focused;
                        focusIndex = ++cur >= length ? 0 : cur;
                    }
                    break;
                case "ArrowUp":
                    {
                        let cur = this._menuItems.focused; // : length;
                        focusIndex = --cur < 0 ? length - 1 : cur;
                    }
                    break;
                case "Home":
                    focusIndex = 0;
                    break;
                case "End":
                    focusIndex = length - 1;
                    break;
                default:
                    break;
            }
        }
        if (focusIndex != null) {
            e.preventDefault();
            this.focusMenuItemByIndex(focusIndex);
        }
    }
    gotKeyDown(e) {
        // if (this.$isPending) {return;} // pending event disables gotKeyDown so case impossible
        if (this._opts.allowNewValue && e.key === "Enter" && !this._focusedMenuItem) {
            this.setValue(this.parseInput(this.$refInput.value), 3 /* SetValueReasons.userInput */);
            this._opts.multiple && e.preventDefault(); // prevent closing by keydown
        }
        !e.defaultPrevented && super.gotKeyDown(e);
    }
    /** Called to filter menuItems (on input change etc.) */
    filterMenuItems() {
        const rawV = this.$refInput.value;
        let v = rawV;
        if (this._opts.multiple) {
            v = rawV.substring(rawV.lastIndexOf(",") + 1, rawV.length);
        }
        v = v.trim().toLowerCase();
        const filtered = [];
        this._menuItems.all.forEach((li, i) => {
            const isOk = this.#ctr.$filterMenuItem.call(this, li._text, li._value, v, rawV);
            isOk && filtered.push(i);
            this.filterMenuItem(li, !isOk);
        });
        const hasFiltered = filtered.length !== this._menuItems.all.length;
        this._menuItems.filtered = hasFiltered ? filtered : undefined;
        const hasVisible = filtered.length !== 0;
        this.renderMenuNoItems(this.$refPopup, hasVisible);
        hasVisible && rawV !== "" && !this._opts.allowNewValue && this.focusMenuItemByIndex(0);
    }
    /** Called on showMenu when user opened it without input-change */
    clearFilterMenuItems() {
        this._menuItems.all.forEach((li) => this.filterMenuItem(li, false)); // reset styles after filtering
        delete this._menuItems.filtered;
        const hasVisible = this._menuItems.all.length !== 0;
        this.renderMenuNoItems(this.$refPopup, hasVisible); // remove NoItems
    }
    /** Called to hide/show menuItem */
    filterMenuItem(el, hide) {
        el.style.display = hide ? "none" : "";
    }
    /** Returns if possible to de-select menu items when input is empty */
    get canClearSelection() {
        return !this.$refInput.value && /*! this.$isRequired && */ !this._opts.multiple;
    }
    gotBeforeInput(e) {
        super.gotBeforeInput(e);
        // handle delete-behavior
        if (this._opts.multiple /* && this.$refInput.value.endsWith(", ") */) {
            let isDelBack = false;
            /* prettier-ignore */
            switch (e.inputType) {
                case "deleteContentBackward":
                    isDelBack = true;
                    break;
                case "deleteContentForward":
                    isDelBack = false;
                    break;
                default: return;
            }
            const pos = this.$refInput.selectionStart || 0;
            let pos2 = this.$refInput.selectionEnd || 0;
            const str = this.$refInput.value;
            const isItemPart = pos < str.lastIndexOf(", ") + (isDelBack ? 3 : 2);
            if (!isItemPart) {
                return;
            }
            const delimitLength = 2; // length of ", "
            const char = ",".charCodeAt(0);
            if (isDelBack && !pos) {
                return;
            }
            const start = isDelBack ? pos - 1 : pos;
            let i = start;
            for (; i !== 0; --i) {
                if (str.charCodeAt(i) === char) {
                    if (!isDelBack || start - i >= delimitLength) {
                        i += delimitLength;
                        // case 1: // "Item 1, Item 2, |" => "Item 1, |Item 2, "
                        // case 2: // "Item 1,| Item 2, " => "|Item 1, Item 2, "
                        break;
                    }
                }
            }
            if (pos === pos2) {
                pos2 = i;
            }
            let end = str.indexOf(",", pos2);
            end = end === -1 ? str.length : end + delimitLength;
            this.$refInput.setSelectionRange(i, end); // select whole items(-s) so input event delete everything itself
            // e.preventDefault(); // uncomment to check behavior manually
        }
    }
    /** Called on input change to update menu if possible */
    tryUpdateMenu() {
        this.$isOpened && this.filterMenuItems();
        this.canClearSelection && this.selectMenuItem(null); // allow user clear value without pressing Enter
    }
    clearValue() {
        super.clearValue();
        this.tryUpdateMenu();
    }
    gotInput(e) {
        this.$isOpened && this.focusMenuItem(null); // reset virtual focus: // WARN it's not good enough when this._opts.multiple
        super.gotInput(e, true); // prevent ordinary value change
        // user can append item by ',' at the end or remove immediately
        if (this._opts.multiple && this.canParseInput(this.$refInput.value)) {
            const str = this.$refInput.value;
            const isDeleted = e.inputType.startsWith("deleteContent");
            const pos = this.$refInput.selectionStart || 0;
            if (pos !== str.length && !isDeleted) {
                this.declineInput(str.length); // don't allow to type text in the middle
            }
            else {
                const iEnd = str.lastIndexOf(",");
                const strVal = str.substring(0, iEnd); // allow user filter via typing in the end of input => 'Item1, Item2,|'
                const next = this.parseInput(strVal);
                // case 1: user removes 'Item1,| Item2|' - setValue [Item1]
                // case 2: user adds `Item1,| Item2,| -- setValue [Item1, Item2]
                const nextLn = (next || []).length;
                const prevLn = (this.$value || []).length;
                const isChanged = nextLn !== prevLn;
                if (isChanged) {
                    this.setValue(next, 3 /* SetValueReasons.userInput */, true);
                    this.setInputValueDirect(this.valueToInput(next) + str.substring(iEnd + 1).trimStart(), 3 /* SetValueReasons.userInput */); // add search-part after value change
                    nextLn < prevLn && this.$refInput.setSelectionRange(pos, pos); // restore caret only if items removed
                }
            }
        }
        this.tryUpdateMenu();
    }
    gotFocus(e) {
        const r = super.gotFocus(e);
        if (this._opts.multiple) {
            if (!this.$isDisabled && !this.$isReadOnly && !this._opts.readOnlyInput && this.$refInput.value) {
                this.setInputValueDirect(`${this.$refInput.value}, `, 3 /* SetValueReasons.userInput */); // add delimiter at the end
            }
        }
        return r;
    }
    gotFocusLost() {
        this._opts.multiple &&
            this.setInputValueDirect(this.$refInput.value.replace(/, *$/, ""), 3 /* SetValueReasons.userInput */); // remove delimiter
        if (this._opts.allowNewValue) {
            this.setValue(this.parseInput(this.$refInput.value), 3 /* SetValueReasons.userInput */); // to allow user left value without pressing Enter
        }
        else {
            this.resetInputValue(); // to update/rollback input according to result
        }
        super.gotFocusLost();
    }
    setValue(v, reason, skipInput = false) {
        const r = super.setValue(v, reason, skipInput);
        r && reason !== 4 /* SetValueReasons.userSelect */ && this.selectMenuItemByValue(v);
        return r;
    }
    removePopup() {
        super.removePopup();
        this._menuItems = undefined;
    }
}
customElements.define(tagName, WUPSelectControl);
// WARN Chrome touchscreen simulation issue: touch on label>strong fires click on input - the issue only in simulation
// WARN label for="" in Chrome sometimes enables autosuggestion - need to remove it for all controls - need to double-check
// NiceToHave: add support custom items rendering when it's already appended to DOM like it works with dropdown
// NiceToHave: option to allow autoselect item without pressing Enter: option: $autoComplete + aria-autocomplete: true => https://www.w3.org/WAI/ARIA/apg/patterns/combobox/examples/combobox-autocomplete-both/
// NiceToHave: color differently text-chunk that matches in menu
// NiceToHave: for allowNewValue add at the end of menu `[text] (New option)` like it works in JIRA. `label` dropdown on ticket
