// import WUPBaseElement from "../baseElement";
import WUPBaseElement from "../baseElement";
import { WUPcssScrollSmall } from "../styles";
let isFirst = true;
/** Represents contenteditable element with custom input props as value, select etc. */
export default class WUPTextareaInput extends HTMLElement {
    /** Returns this.constructor // watch-fix: https://github.com/Microsoft/TypeScript/issues/3841#issuecomment-337560146 */
    #ctr = this.constructor;
    static $use() {
        // it's for sideEffects component self-registered
    }
    static get $style() {
        return `:host {
          display: inline-block; ${ /* it removes extra space below */""}
          cursor: text;
          white-space: pre-wrap;
          word-break: break-word;
          overflow-wrap: break-word;
          overflow: auto;
          margin: 0; padding: 0;
        }
        ${WUPcssScrollSmall(":host")}`;
    }
    constructor() {
        super();
        if (isFirst) {
            WUPBaseElement.$refStyle.append(this.#ctr.$style.replace(/:host/g, `${this.tagName}`));
            isFirst = false;
        }
    }
    $options = {};
    #isInit = true;
    connectedCallback() {
        if (this.#isInit) {
            this.setAttribute("contenteditable", "true");
            this.setAttribute("role", "textbox");
            this.setAttribute("aria-multiline", "true");
            this.#isInit = false;
            this.oninput = () => {
                this._cached = undefined;
            };
        }
    }
    /** Makes the selection equal to the current object. */
    select() {
        const range = document.createRange();
        range.selectNodeContents(this);
        const sel = window.getSelection();
        sel.removeAllRanges();
        sel.addRange(range);
    }
    /* Get/set readonly */
    get readOnly() {
        return this.hasAttribute("aria-readonly");
    }
    set readOnly(v) {
        WUPBaseElement.prototype.setAttr.call(this, "aria-readonly", v);
    }
    _cached;
    /** Get/set innerHTML (br converted into '\n') */
    get value() {
        if (this._cached == null) {
            this._cached = this.innerHTML
                // .replace(/\n $/, "\n")
                .replace(/<div><br><\/div>/g, "\n") // Chrome newLine
                .replace(/<br>|<div>/g, "\n")
                .replace(/<\/div>/g, "")
                .replace(/^&nbsp;/, "");
        }
        return this._cached;
    }
    set value(v) {
        this._cached = undefined;
        this.innerHTML = v;
    }
    /** Sets the start and end positions of a selection in a text field.
     * @param start The offset into the text field for the start of the selection.
     * @param end The offset into the text field for the end of the selection. */
    setSelectionRange(start, end) {
        this.selection = { start: start || 0, end: end || 0 };
    }
    /** Gets or sets the starting position or offset of a text selection. */
    get selectionStart() {
        return this.selection?.start ?? null;
    }
    set selectionStart(v) {
        this.selection = { start: v || 0, end: this.selection?.end };
    }
    /** Gets or sets the end position or offset of a text selection. */
    get selectionEnd() {
        return this.selection?.end ?? null;
    }
    set selectionEnd(v) {
        this.selection = { start: this.selection?.start, end: v || 0 };
    }
    /** Fix for contenteditable (it doesn't contain selectionStart & selectionEnd props) */
    get selection() {
        if (document.activeElement !== this) {
            return null;
        }
        const sel = window.getSelection();
        if (!sel || sel.rangeCount < 1) {
            return null;
        }
        const range = sel.getRangeAt(0);
        const pre = range.cloneRange();
        pre.selectNodeContents(this);
        pre.setEnd(range.startContainer, range.startOffset);
        const start = pre.toString().length;
        return {
            start,
            end: start + range.toString().length,
        };
    }
    set selection(sel) {
        if (document.activeElement !== this) {
            return;
        }
        let charIndex = 0;
        const range = document.createRange();
        range.setStart(this, 0);
        range.collapse(true);
        const nodeStack = [this];
        let node;
        let foundStart = false;
        let stop = false;
        // from https://stackoverflow.com/questions/13949059/persisting-the-changes-of-range-objects-after-selection-in-html/13950376#13950376
        /* istanbul ignore else */
        if (sel) {
            // eslint-disable-next-line no-cond-assign
            while (!stop && (node = nodeStack.pop())) {
                if (node.nodeType === Node.TEXT_NODE) {
                    const nextCharIndex = charIndex + node.length;
                    /* istanbul ignore else */
                    if (!foundStart && sel.start >= charIndex && sel.start <= nextCharIndex) {
                        range.setStart(node, sel.start - charIndex);
                        foundStart = true;
                    }
                    /* istanbul ignore else */
                    if (foundStart && sel.end >= charIndex && sel.end <= nextCharIndex) {
                        range.setEnd(node, sel.end - charIndex);
                        stop = true;
                    }
                    charIndex = nextCharIndex;
                }
                else {
                    let i = node.childNodes.length;
                    while (i--) {
                        nodeStack.push(node.childNodes[i]);
                    }
                }
            }
        }
        const s = window.getSelection();
        s.removeAllRanges();
        s.addRange(range);
    }
}
customElements.define("wup-areainput", WUPTextareaInput);
