import WUPBaseModal from "./baseModal";
import focusFirst from "./helpers/focusFirst";
import { WUPcssButton } from "./styles";
export var ModalOpenCases;
(function (ModalOpenCases) {
    /** When $open() is called programmatically */
    ModalOpenCases[ModalOpenCases["onManualCall"] = 0] = "onManualCall";
    /** On init (when appended to layout) */
    ModalOpenCases[ModalOpenCases["onInit"] = 1] = "onInit";
    /** When click on target @see {@link WUP.Modal.Options.target} */
    ModalOpenCases[ModalOpenCases["onTargetClick"] = 2] = "onTargetClick";
})(ModalOpenCases || (ModalOpenCases = {}));
export var ModalCloseCases;
(function (ModalCloseCases) {
    /** When $close() is called programmatically */
    ModalCloseCases[ModalCloseCases["onManualCall"] = 0] = "onManualCall";
    /** When was click on button[close] */
    ModalCloseCases[ModalCloseCases["onCloseClick"] = 1] = "onCloseClick";
    /** When was click outside modal */
    ModalCloseCases[ModalCloseCases["onOutsideClick"] = 2] = "onOutsideClick";
    /** When user pressed Escape button */
    ModalCloseCases[ModalCloseCases["onPressEsc"] = 3] = "onPressEsc";
    /* When on successful wup-form.$onSubmitEnd: @see {@link WUP.Form.EventMap.$submitEnd} */
    ModalCloseCases[ModalCloseCases["onSubmitEnd"] = 4] = "onSubmitEnd";
})(ModalCloseCases || (ModalCloseCases = {}));
const tagName = "wup-modal";
const attrConfirm = "w-confirm";
/** Modal element
 * @example
 * JS/TS
 * ```js
 * WUPModalElement.$defaults.placement = ...;
 *
 * const el = document.createElement('wup-modal');
 * el.textContent = ...;
 * document.body.append(el);
 * el.$onClose = () => el.remove(); // self-remove after close
 *```
 * HTML
 * ```html
 * <wup-modal>Some content here</wup-modal>
 * ```
 * @tutorial Troubleshooting known issues:
 * * for very long content with 1st focusable item at the bottom it won't be visible because user must see top of the modal at first
 * To fix the issue set `<h2 tabindex="-1">...</h2>` and hide focus frame via styles OR disable `$options.autoFocus`
 * * accessibility: NVDA reads modal content twice. To fix follow the recommendations: https://github.com/nvaccess/nvda/issues/8971 */
export default class WUPModalElement extends WUPBaseModal {
    #ctr = this.constructor;
    /** Call it to enable attribute [w-confirm] for buttons
     * @tutorial Rules
     * * to override default render: redefine `WUPModalElement.prototype.gotRenderConfirm` method OR use `onRender` callback; @see {@link WUPModalElement.$showConfirm}
     * @example
     * ```html
     * <button w-confirm="Do you want to do it?">
     *   I will fire click event only when confirmButton will be pressed in the confirm-modal
     * </button>
     * ``` */
    static $useConfirmHook(opts) {
        document.addEventListener("click", (e) => {
            if (e.button || window.__wupFixCycleClick) {
                return;
            }
            const btn = WUPModalElement.prototype.findParent.call(e.target, (el) => el.hasAttribute(attrConfirm), {
                bubbleCount: 5, // no more 5 parent iterations
            });
            const txt = btn?.getAttribute(attrConfirm);
            if (txt) {
                e.stopPropagation();
                this.$showConfirm({
                    question: txt,
                    ...opts,
                    onRender: (m) => {
                        btn.$onRenderModal?.call(btn, m);
                        opts?.onRender?.call(this, m);
                    },
                }).then((isOk) => {
                    isOk && btn.dispatchEvent(new MouseEvent("click", { bubbles: true }));
                });
            }
        }, { capture: true });
    }
    /** Show confirm modal and return Promise(true) if user pressed button[data-close=confirm] and before modal is closed itself
     * @tutorial Troubleshooting
     * * ConfirmModal ignores option `placement` if it overflow existed modal
     * * On close modal self-removed */
    static async $showConfirm(opts) {
        // init
        const me = document.createElement("wup-modal");
        opts?.className && me.classList.add(opts.className);
        me.$options.selfRemove = true;
        Object.assign(me.$options, opts?.defaults);
        me.gotRenderConfirm(opts?.question || "");
        // append to root
        document.body.appendChild(me);
        // call callback to allow user re-define it
        opts?.onRender?.call(this, me);
        setTimeout(() => {
            if (me._mid && me._mid > 1) {
                const p = me._openedModals[me._mid - 2];
                // case impossible because _openedModal must contain only opened modals
                // if (!p.$isOpened || p.$isClosing) {
                //   return; // it should work only if prev modal is opened
                // }
                // setup position according to options: center in modal if `overflow` OR in center of screen if `replace` +  default left/right
                if (me._opts.replace) {
                    switch (me._opts.placement) {
                        case "left":
                        case "right":
                            me.setAttribute("w-placement", "center");
                            break;
                        default:
                            break;
                    }
                    return;
                }
                // position in center of parent modal
                me.setAttribute("w-placement", "center");
                const prev = { x: -1, y: -1 };
                const goCenter = () => {
                    if (!me.$isOpened || me.$isClosing || !p.isConnected) {
                        return;
                    }
                    const r = p.getBoundingClientRect();
                    const rme = me.getBoundingClientRect();
                    const x = Math.max(0, Math.round(r.x + (r.width - rme.width) / 2)); // WARN: expected that modal never bigger than viewport
                    const y = Math.max(0, Math.round(r.y + (r.height - rme.height) / 2));
                    if (prev.x !== x || prev.y !== y) {
                        prev.x = x;
                        prev.y = y;
                        me.style.margin = `${y}px 0 0 ${x}px`; // center of parent modal
                    }
                    window.requestAnimationFrame(goCenter);
                };
                goCenter();
            }
        }); // timeout to wait when this.openedModals[] changed
        const btnConfirm = me.querySelector("[data-close=confirm]");
        let isConfirmed = false;
        if (btnConfirm) {
            btnConfirm.onclick = (ev) => {
                isConfirmed = true;
                setTimeout(() => (isConfirmed = false), 1);
                window.__wupFixCycleClick = true;
                setTimeout(() => delete window.__wupFixCycleClick); // to prevent false opening next confirm window
                me.goClose(1 /* ModalCloseCases.onCloseClick */, ev);
            };
        }
        else {
            me.throwError("button[data-close=confirm] missed");
        }
        return new Promise((res) => {
            // WARN: expected that $willClose isn't prevented otherwise need to add logic for checking this
            me.addEventListener("$willClose", () => res(isConfirmed), { once: true, passive: true });
        });
    }
    // static get observedOptions(): Array<keyof WUP.Modal.Options> { return []; }
    // static get observedAttributes(): Array<string> { return []; }
    /** Default class used for fade - blurring background for main content
     * @defaultValue "wup-modal-fade" */
    static $classFade = "wup-modal-fade";
    /** Default class that appended to body when modal opened (required to hide body scroll)
     * @defaultValue "wup-modal-open" */
    static $classOpened = "wup-modal-open";
    // WARN: modal-anim shouldn't affect on animation of nested
    static get $styleRoot() {
        return `:root {
        --modal-anim-t: 400ms;
        --modal-text: inherit;
        --modal-bg: #fff;
        --modal-fade: #0007;
        --modal-margin: 2em;
      }
      [wupdark] {
        --modal-bg: #222a36;
      }`;
    }
    static get $style() {
        return `${super.$style}
      :host {
        z-index: 9002;
        width: 100%;
        max-width: 600px;
        min-height: 150px;
        color: var(--modal-text);
        background: var(--modal-bg);
        border: 1px solid #0002;
        user-select: none;
        pointer-events: none;
        touch-action: none;
        outline: none;
      }
      :host[open] {
        user-select: initial;
        pointer-events: initial;
        touch-action: initial;
       }
      :host[w-placement="top"] {
        top:0;left:0;right:0;
        margin: var(--modal-margin);
        margin-left: auto;
        margin-right: auto;
        max-height: calc(100% - var(--modal-margin) * 2);
        transform: translateY(-50%);
      }
      :host[w-placement="center"] {
        top:0;left:0;right:0;bottom:0;
        margin: auto;
        height: fit-content${
        /* This is tricky rule. If it will buggy need to rewrite centering logic via js-core like it works with popup  */ ""};
        max-height: calc(100% - var(--modal-margin) * 2);
        transform: translateY(-150%);
      }
      :host[w-placement="right"] {
        right:0;top:0;bottom:0;
        border-radius: 0;
        transform: translateX(100%);
      }
      :host[w-placement="left"] {
        left:0;top:0;bottom:0;
        border-radius: 0;
        transform: translateX(-100%);
      }
      :host[show] {
        transform: none;
      }
      :host[hide] {
        opacity: 0!important;
        pointer-events: none!important;
        user-select: none!important;
        touch-action: none!important;
      }
      @media (max-width: 600px) {
        :host {
          --modal-margin: 0px;
          border-radius: 0;
        }
      }
      .${this.$classFade} {
         --modal-anim: var(--modal-anim-t) cubic-bezier(0, 0, 0.2, 1) 0ms;
        z-index: 9000;
        display: block;
        position: fixed;
        top:0;left:0;right:0;bottom:0;
        background: var(--modal-fade);
        opacity: 0;
        pointer-events: none;
        touch-action: none;
      }
      .${this.$classFade}[show] {
         pointer-events: initial;
         touch-action: initial;
         opacity: 1;
      }
      @media not all and (prefers-reduced-motion) {
       .${this.$classFade} {
          transition: opacity var(--modal-anim), transform var(--modal-anim);
        }
      }
      :host > button[close] {
        --icon-img: var(--wup-icon-cross);
        position: absolute;
        right: 0;
        margin: -0.2em 1em 0;
      }
      :host h2 {
        margin: 0 0 1em;
        padding-right: 1.8em;
      }
      :host wup-form {
        max-width: initial;
      }
      :host wup-form button[type=submit] {
        margin-bottom: 0;
        margin-left: auto;
      }
      :host footer {
        display: flex;
        gap: calc(var(--base-margin) / 2);
        margin-top: var(--base-margin);
        justify-content: flex-end;
      }
      ${WUPcssButton(":host footer>button")}
      :host footer>button[type] {
         margin: 0;
         min-width: 7em;
      }
      :host footer>button[data-close=modal] {
         background: var(--base-btn2-bg);
         color: var(--base-btn2-text);
      }`;
    }
    static $defaults = {
        target: null,
        placement: "center",
        autoFocus: true,
        autoClose: true,
        selfRemove: false,
        replace: false,
        confirmUnsaved: true,
    };
    static get mappedAttributes() {
        const m = super.mappedAttributes;
        m.target = { type: 6 /* AttributeTypes.selector */ };
        return m;
    }
    /** Reference to fade element */
    $refFade;
    /** Reference to button[close] */
    $refClose;
    gotRender() {
        /** empty because component is hidden by default and need to focus on open-phase */
    }
    /** Override it to change default render for modalConfirm */
    gotRenderConfirm(headerContent) {
        this.innerHTML = `<h2>${headerContent}</h2>
<footer>
  <button type='button' data-close='modal'>${__wupln("Cancel", "content")}</button>
  <button type='button' data-close='confirm'>${__wupln("Confirm", "content")}</button>
</footer>`;
    }
    /** Related button that need to listen for click event to open modal */
    _target;
    /** Click event related to _target */
    _targetClick;
    gotChanges(propsChanged) {
        super.gotChanges(propsChanged);
        const trg = this._opts.target;
        const isTrgChange = this._target !== trg;
        if (isTrgChange) {
            this._target?.removeEventListener("click", this._targetClick);
            this._target = trg;
        }
        if (!trg) {
            !propsChanged && this.goOpen(1 /* ModalOpenCases.onInit */, null); // call here to wait for options before opening
        }
        else if (isTrgChange) {
            this._targetClick = (e) => {
                // timeout required to prevent opening from bubbled-events
                setTimeout(() => !e.defaultPrevented && this.goOpen(2 /* ModalOpenCases.onTargetClick */, e));
            };
            trg.addEventListener("click", this._targetClick, { passive: true });
        }
    }
    /** Id of last focused item on item itself */
    _lastFocused;
    /** Number of opened modal */
    _mid;
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    gotOpen(openCase, e) {
        // store id of focused element
        const ael = document.activeElement;
        this._lastFocused = ael?.id || ael;
        // setup Accessibility attrs
        this.tabIndex = -1; // WA: to allow scroll modal - possible that tabindex: -1 doesn't allow tabbing inside
        this.role = "dialog"; // possible alert-dialog for Confirm modal
        this.setAttribute("aria-modal", true); // WARN for old readers it doesn't work and need set aria-hidden to all content around modal
        this._target?.$onRenderModal?.call(this._target, this);
        const header = this.querySelector("h1,h2,h3,h4,h5,h6,[role=heading]");
        if (!header) {
            console.warn("WA: header missed. Add <h2>..<h6> or role='heading' to modal content");
        }
        else {
            header.id ||= this.#ctr.$uniqueId;
            this.setAttribute("aria-labelledby", header.id); // issue: NVDA reads content twice if modal has aria-labelledby: https://github.com/nvaccess/nvda/issues/8971
        }
        // init button[close]
        this.$refClose ??= document.createElement("button");
        this.$refClose.type = "button";
        this.$refClose.setAttribute("aria-label", __wupln("close", "aria"));
        this.$refClose.setAttribute(this.#ctr.classNameBtnIcon, "");
        this.$refClose.setAttribute("close", "");
        this.$refClose.setAttribute("data-close", "modal");
        this.prepend(this.$refClose);
        // init fade
        this._openedModals.push(this);
        this._mid = this._openedModals.length;
        if (this._mid === 1) {
            this.$refFade ??= document.body.appendChild(document.createElement("div"));
            setTimeout(() => this.$refFade?.setAttribute("show", "")); // timeout to allow animation works
        }
        else {
            const p = this._openedModals[this._mid - 2];
            const isReplace = this._opts.replace;
            this.$refFade ??= (isReplace ? document.body : p).appendChild(document.createElement("div")); // append to parent to hide modal parent
            this.$refFade.setAttribute("show", ""); // WARN without timeout to show immediately
            p.$refFade.style.display = "none";
            isReplace && p.setAttribute("hide", ""); // hide such modal
        }
        this.$refFade.className = this.#ctr.$classFade;
        // apply open styles
        this.setAttribute("w-placement", this._opts.placement);
        document.body.classList.add(this.#ctr.$classOpened); // to maybe hide scroll-bars
        // listen for close-events
        this.$refFade.onclick = (ev) => this.goClose(2 /* ModalCloseCases.onOutsideClick */, ev);
        this.appendEvent(this, "click", (ev) => this.gotClick(ev));
        this.appendEvent(this, "keydown", (ev) => this.gotKeyDown(ev), { passive: false });
        // @ts-expect-error - TS isn't good enough for looking for types
        this.appendEvent(this, "$submitEnd", (sev) => {
            sev.detail.success && this.goClose(4 /* ModalCloseCases.onSubmitEnd */, sev);
        });
        // this.appendEvent(this, "focusout", (ev) => this.gotFocusOut(ev), { passive: false });
        this._opts.autoFocus ? this.focusAny() : this.focus(); // bug: FF doesn't adjust suggest-popup according to animation
    }
    goClose(closeCase, ev, immediately) {
        if (this._whenClose) {
            return this._whenClose;
        }
        if (!this._whenOpen) {
            return Promise.resolve(true); // possible when on init $close is called
        }
        if (this._opts.confirmUnsaved && closeCase !== 4 /* ModalCloseCases.onSubmitEnd */) {
            const f = this.querySelector("wup-form");
            if (f?.$isChanged && !f.$options.autoStore) {
                return this.#ctr
                    .$showConfirm({ question: __wupln("You have unsaved changes.\nWould you like to skip it?", "content") })
                    .then((isOk) => (isOk ? super.goClose(closeCase, ev, immediately) : false));
            }
        }
        return super.goClose(closeCase, ev, immediately);
    }
    // eslint-disable-next-line @typescript-eslint/no-unused-vars
    gotClose(closeCase, ev) {
        // remove events
        this.$refFade.onclick = null;
        this.disposeEvents(); // remove only events WARN: don't call this.dispose()!
        // apply animation
        if (this._mid === 1) {
            this.$refFade.removeAttribute("show"); // animation if opened single modal
            const b = document.body;
            b.classList.remove(this.#ctr.$classOpened); // testCase: on modal.remove everything must returned to prev state
            !b.className && b.removeAttribute("class");
        }
        else {
            this.$refFade.remove(); // immediately hide if opened 2+ modals
            this.$refFade = undefined;
            const p = this._openedModals[this._mid - 2];
            p.$refFade.style.display = "";
            this.removeEmptyStyle.call(p.$refFade);
            p.removeAttribute("hide");
        }
        this._openedModals.splice(this._mid - 1, 1); // self remove from array
        this.focusBack();
    }
    /** Called when modal handles click to check if was close-click */
    gotClick(e) {
        const t = e.target;
        if (e.defaultPrevented || t === this || e.button) {
            return;
        }
        const all = this.querySelectorAll("[data-close=modal]").values(); // allow to use any button with attr to close modal
        // eslint-disable-next-line no-restricted-syntax
        for (const el of all) {
            if (this.itsMe.call(el, t)) {
                this._mid !== 1 && e.stopPropagation(); // prevent closing other modals
                this.goClose(1 /* ModalCloseCases.onCloseClick */, e);
                break;
            }
        }
    }
    // WARN: the focus block work better because not handles keyboard but looks for focus behavior but it's buggy in FF and maybe other browsers since focusOut can be outside the modal
    // /** Required for tab-cycle behavior */
    // _willFocusPrev?: boolean;
    // /** Called on focusout event */
    // gotFocusOut(e: FocusEvent): void {
    //   const t = e.relatedTarget;
    //   const isOut = !t || !this.includes(t); // WARN: includes can be wrong for absolute items ???
    //   const isFocusLast = this._willFocusPrev;
    //   isOut && isFocusLast != null && focusFirst(this, { isFocusLast }); // bug: FF doesn't remove focus-style on btn[close] when focus goes outside
    //   delete this._willFocusPrev;
    //   console.warn("blur", { isFocusLast, relatedTarget: t, hasFocus: document.hasFocus() });
    //   // WARN: browser can ignore focus action and move focus to the tab panel - it's ok and even better than whole cycling
    // }
    /** Called on keydown event */
    gotKeyDown(e) {
        if (e.altKey || e.ctrlKey || e.metaKey) {
            return;
        }
        switch (e.key) {
            case "Escape":
                if (!e.shiftKey && !e.defaultPrevented) {
                    e.preventDefault();
                    this.goClose(3 /* ModalCloseCases.onPressEsc */, e);
                }
                break;
            case "Tab": {
                let isFocusLast;
                if (e.shiftKey) {
                    const isFirst = document.activeElement === this.querySelector(focusFirst.$selector);
                    if (isFirst) {
                        isFocusLast = true; // to cycle focus-behavior
                    }
                }
                else {
                    const all = this.querySelectorAll(focusFirst.$selector);
                    const isLast = document.activeElement === all.item(all.length - 1);
                    if (isLast) {
                        isFocusLast = false; // to cycle focus-behavior
                    }
                }
                if (isFocusLast != null) {
                    e.preventDefault(); // prevent default focus behavior
                    focusFirst(this, { isFocusLast });
                }
                break;
            }
            default:
                break;
        }
    }
    /** Called on close to return focus to previously focused item */
    focusBack() {
        let el;
        if (typeof this._lastFocused === "string") {
            el = document.getElementById(this._lastFocused);
        }
        else {
            el = this._lastFocused;
        }
        if (el && el.focus && el.isConnected) {
            el.focus();
        }
        else {
            this.throwError("Impossible to return focus back: element is missed. Before opening modal set 'id' to focused element");
            /* istanbul ignore next */
            document.activeElement?.blur?.();
        }
    }
    /** Focus any content excluding button[close] if possible */
    focusAny() {
        if (this.$isOpened) {
            this.$refClose.disabled = true;
            const isOk = this.focus();
            this.$refClose.disabled = false;
            !isOk && this.$refClose.focus();
            // if (document.activeElement !== this.$refClose) {
            //   this.tabIndex = 0;
            //   HTMLElement.prototype.focus.call(this);
            // }
        }
    }
    resetState() {
        super.resetState();
        this.$refFade?.remove();
        this.$refFade = undefined;
        this.isConnected && this._opts.selfRemove && this.remove();
    }
    dispose() {
        const bd = document.body;
        bd.classList.remove(this.#ctr.$classOpened); // testCase: on modal.remove > everything must returned to prev state
        !bd.className && bd.removeAttribute("class");
        this._target?.removeEventListener("click", this._targetClick);
        const i = this._openedModals.findIndex((m) => this === m);
        if (i > -1) {
            this._openedModals.splice(i, 1);
        }
        super.dispose();
    }
    /** Singleton array with opened modals */
    get _openedModals() {
        return __openedModals;
    }
}
const __openedModals = [];
customElements.define(tagName, WUPModalElement);
// NiceToHave: handle Ctrl+S, Meta+S for submit & close ???
